import logging
import sys
from contextlib import contextmanager
from pathlib import Path

from racetrack_client.log.debug import debug_endpoints_enabled

DEBUG_LOGS_PATH = '/tmp/ikp-rt/lifecycle.log'


@contextmanager
def save_logs():
    """Within scope of this context manager all stdout & stderr are written to a log file"""
    if not debug_endpoints_enabled():
        yield
        return

    stdout_original = sys.stdout
    stderr_original = sys.stderr
    try:
        Path(DEBUG_LOGS_PATH).parent.mkdir(parents=True, exist_ok=True)
        with open(DEBUG_LOGS_PATH, 'w') as file_stream:
            sys.stdout = ForkStreamer(stdout_original, file_stream)
            sys.stderr = ForkStreamer(stderr_original, file_stream)
            yield
    finally:
        sys.stdout = stdout_original
        sys.stderr = stderr_original


def read_logs() -> str:
    log_path = Path(DEBUG_LOGS_PATH)
    if not log_path.is_file():
        logging.warning(f"Log file doesn't exist: {log_path}")
        return ''
    return log_path.read_text()


class ForkStreamer:
    """A stream bifurcating output to two output streams."""

    def __init__(self, stream1, stream2):
        self.stream1 = stream1
        self.stream2 = stream2

    def write(self, output):
        self.stream1.write(output)
        self.stream2.write(output)

    def flush(self):
        self.stream1.flush()
        self.stream2.flush()



import logging
from typing import Optional

LOG_FORMAT = '[%(asctime)s] %(levelname)s: %(message)s'
LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'


def init_logs():
    """Run logging initialization (before loading log level)"""
    logging.basicConfig(format=LOG_FORMAT, level=logging.INFO, datefmt=LOG_DATE_FORMAT)


def configure_logs(verbosity: int = 0, log_level: Optional[str] = None):
    """Configure root logger with a log level"""
    if log_level is None:
        if verbosity > 0:
            level = logging.DEBUG
        else:
            level = logging.INFO
    else:
        level = _get_logging_level(log_level)
    logging.basicConfig(format=LOG_FORMAT, level=level, datefmt=LOG_DATE_FORMAT)


def _get_logging_level(str_level: str) -> int:
    return {
        'debug': logging.DEBUG,
        'info': logging.INFO,
        'warn': logging.WARNING,
        'warning': logging.WARNING,
        'error': logging.ERROR,
        'crit': logging.CRITICAL,
        'off': logging.NOTSET,
    }[str_level.lower()]



from dataclasses import dataclass, field
from typing import Dict


@dataclass
class ClientConfig:
    """Global options for a local client"""

    # default URL of Racetrack API server (Lifecycle URL)
    lifecycle_url: str = 'http://127.0.0.1:7002'

    # Git auth credentials set for particular repositories
    git_credentials: Dict[str, 'Credentials'] = field(default_factory=dict)

    # Racetrack URL aliases: alias name -> full URL to Lifecycle API
    lifecycle_url_aliases: Dict[str, str] = field(default_factory=dict)


@dataclass
class Credentials:
    username: str
    password: str



from typing import Optional

from racetrack_client.client_config.client_config import ClientConfig


def resolve_lifecycle_url(client_config: ClientConfig, lifecycle_name: Optional[str]) -> str:
    """
    Get full URL of Lifecycle based on given short name.
    If aliased name is found, resolve it to full URL it's pointing to.
    If the name was not provided, return default URL.
    """
    if lifecycle_name is None:
        return client_config.lifecycle_url
    if lifecycle_name not in client_config.lifecycle_url_aliases:
        return lifecycle_name
    return client_config.lifecycle_url_aliases[lifecycle_name]



import logging
import os
from pathlib import Path
from typing import List, Type, TypeVar

import dacite
import yaml

T = TypeVar("T")


def load_config(clazz: Type[T], cast_types: List[Type] = None) -> T:
    """
    Load general configuration from YAML file given in CONFIG_FILE environment var or load default config.
    :param clazz: dataclass config type that should be loaded
    :param cast_types: list of types that should be called to create an instance of a field type from the input data
    :return: configuration object of given "clazz" type
    """
    config_file_path = os.environ.get('CONFIG_FILE')
    if not config_file_path:
        logging.warning('CONFIG_FILE unspecified, loading default config')
        return clazz()

    path = Path(config_file_path)
    if not path.is_file():
        raise FileNotFoundError(f'config file {config_file_path} doesnt exist')

    try:
        with path.open() as file:
            config_dict = yaml.load(file, Loader=yaml.FullLoader)
            config = dacite.from_dict(
                data_class=clazz,
                data=config_dict,
                config=dacite.Config(cast=cast_types or []),
            )

            logging.info(f'config loaded from {config_file_path}: {config}')
            return config
    except Exception as e:
        raise RuntimeError('loading config failed') from e



import dataclasses
import logging
from typing import Optional, Dict
from urllib.parse import urlsplit

import backoff
import requests
import urllib3

from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.client_config import ClientConfig, Credentials
from racetrack_client.client_config.io import load_client_config
from racetrack_client.manifest import Manifest
from racetrack_client.manifest.validate import load_validated_manifest
from racetrack_client.utils.request import parse_response

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

DEPLOYMENT_TIMEOUT_SECS = 60 * 60  # 1 hour


class DeploymentError(RuntimeError):
    def __init__(self, cause: BaseException):
        super().__init__()
        self.__cause__ = cause

    def __str__(self):
        return f'deployment error: {self.__cause__}'


def send_deploy_request(
    workdir: str,
    manifest: Optional[Manifest] = None,
    client_config: Optional[ClientConfig] = None,
    lifecycle_url: Optional[str] = None,
):
    """Send request deploying a new Fatman to running LC instance"""
    if client_config is None:
        client_config = load_client_config()
    if manifest is None:
        manifest = load_validated_manifest(workdir)
        logging.debug(f'Manifest loaded: {manifest}')
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    logging.info(f'deploying workspace "{workdir}" (job "{manifest.name}") to {lifecycle_url}')

    _handle_local_repository(manifest, lifecycle_url)
    git_credentials = get_git_credentials(manifest, client_config)

    # see `lifecycle.server.endpoints::setup_api_endpoints::DeployEndpoint` for server-side implementation
    r = requests.post(
        f'{lifecycle_url}/api/v1/deploy',
        json=_get_deploy_request_payload(manifest, git_credentials),
        verify=False,
    )
    response = parse_response(r, 'LC-API deploying error')
    deploy_id = response["id"]
    logging.info(f'job deployment requested: {deploy_id}')

    try:
        fatman_url = _wait_for_deployment_result(lifecycle_url, deploy_id)
    except Exception as e:
        raise DeploymentError(e)

    logging.info(f'Fatman "{manifest.name}" has been deployed. Check out {fatman_url} to access your fatman')


def _get_deploy_request_payload(manifest: Manifest, git_credentials: Optional[Credentials]) -> Dict:
    return {
        "manifest": dataclasses.asdict(manifest),
        "git_credentials": dataclasses.asdict(git_credentials) if git_credentials is not None else None,
    }


@backoff.on_exception(
    backoff.fibo, TimeoutError, max_value=5, max_time=DEPLOYMENT_TIMEOUT_SECS, jitter=None, logger=None
)
def _wait_for_deployment_result(lifecycle_url: str, deploy_id: str) -> str:
    # see `lifecycle.server.endpoints::setup_api_endpoints::DeployIdEndpoint` for server-side implementation
    r = requests.get(
        f'{lifecycle_url}/api/v1/deploy/{deploy_id}',
        verify=False,
    )
    response = parse_response(r, 'LC-API deployment status')
    status = response['status'].lower()
    if status == 'failed':
        raise RuntimeError(response['error'])
    if status == 'done':
        return response.get('fatman', {}).get('pub_url')

    logging.info(f'deployment in progress...')
    raise TimeoutError('Deployment timeout error')


def _handle_local_repository(manifest: Manifest, lifecycle_url: str):
    """Avoid fetching from GitLab when using samples from local project repository"""
    if manifest.git.remote == 'https://rep.erst.dk/git/kubernilla/ikp-rt' and _is_url_localhost(lifecycle_url):
        logging.warning('changing git.remote to "." due to local development')
        manifest.git.remote = '.'


def get_git_credentials(manifest: Manifest, client_config: ClientConfig) -> Optional[Credentials]:
    if manifest.git.remote in client_config.git_credentials:
        return client_config.git_credentials[manifest.git.remote]

    split = urlsplit(manifest.git.remote)

    url = f'{split.scheme}://{split.netloc}{split.path}'
    if url in client_config.git_credentials:
        return client_config.git_credentials[url]

    url = f'{split.scheme}://{split.netloc}'
    if url in client_config.git_credentials:
        return client_config.git_credentials[url]

    url = f'{split.netloc}'
    if url in client_config.git_credentials:
        return client_config.git_credentials[url]

    return None


def _is_url_localhost(url: str) -> bool:
    split = urlsplit(url)
    return split.hostname in {'localhost', '127.0.0.1'}



import io
import logging
import subprocess
import sys
from pathlib import Path
from typing import Optional


def shell(cmd: str, workdir: Optional[Path] = None):
    """
    Run system shell command.
    Print live stdout as it comes (line by line) and capture entire output in case of errors.
    :raises:
        CommandError: in case of non-zero command exit code.
    """
    _run_shell_command(cmd, workdir)


def shell_output(cmd: str, workdir: Optional[Path] = None) -> str:
    """
    Run system shell command and return its output.
    Print live stdout as it comes (line by line) and capture entire output in case of errors.
    """
    captured_stream = _run_shell_command(cmd, workdir, print_stdout=False)
    return captured_stream.getvalue()


def _run_shell_command(cmd: str, workdir: Optional[Path] = None, print_stdout: bool = True) -> io.StringIO:
    logging.info(f'Command: {cmd}')
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, cwd=workdir)
    try:
        # fork command output to stdout & captured buffer
        captured_stream = io.StringIO()
        for line in iter(process.stdout.readline, b''):
            line_str = line.decode()
            if print_stdout:
                sys.stdout.write(line_str)
                sys.stdout.flush()
            captured_stream.write(line_str)

        process.wait()
        if process.returncode != 0:
            stdout = captured_stream.getvalue()
            raise CommandError(cmd, stdout, process.returncode)
        return captured_stream
    except KeyboardInterrupt:
        logging.warning('killing subprocess')
        process.kill()
        raise


class CommandError(RuntimeError):
    def __init__(self, cmd: str, stdout: str, returncode: int):
        super().__init__()
        self.cmd = cmd
        self.stdout = stdout
        self.returncode = returncode

    def __str__(self):
        return f'command failed: {self.cmd}: {self.stdout}'



import logging
from pathlib import Path
from typing import Dict, Optional

import yaml
from racetrack_client.log.context_error import wrap_context
from racetrack_client.manifest.manifest import Manifest
from racetrack_client.utils.dataclass import parse_dict_dataclass, parse_yaml_dataclass

FATMAN_MANIFEST_FILENAME = 'fatman.yaml'


def load_manifest_from_yaml(path: Path) -> Manifest:
    """Load Manifest from a YAML file. Data types validation happens here when parsing YAML."""
    if not path.is_file():
        raise FileNotFoundError(f"manifest file '{path}' doesn't exist")

    with wrap_context('loading YAML'):
        with path.open() as file:
            manifest_dict = yaml.load(file, Loader=yaml.FullLoader)

    return load_manifest_from_dict(manifest_dict)


def load_manifest_from_dict(manifest_dict: Dict) -> Manifest:
    """Return manifest as data class"""
    with wrap_context('parsing manifest data types'):
        return parse_dict_dataclass(manifest_dict, Manifest)


def parse_manifest_or_empty(manifest_yaml: Optional[str]) -> Optional[Manifest]:
    """Parse YAML string as Manifest. In case of error return None"""
    if not manifest_yaml:
        return None
    try:
        return parse_yaml_dataclass(manifest_yaml, Manifest)
    except Exception as e:
        logging.error(f'Fatman Manifest YAML contains syntax error: {e}')
        return None



import logging
from typing import Optional

import requests
import urllib3
from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.io import load_client_config
from racetrack_client.manifest.validate import load_validated_manifest
from racetrack_client.utils.request import parse_response

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def show_logs(workdir: str, lifecycle_url: Optional[str], tail: int):
    """
    Show logs from fatman output
    :param workdir: directory with fatman.yaml manifest
    :param lifecycle_url: URL to Lifecycle API
    :param tail: number of recent lines to show
    """
    client_config = load_client_config()
    manifest = load_validated_manifest(workdir)
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    logging.info(f'Retrieving logs of "{manifest.name}" fatman from {lifecycle_url}...')
    r = requests.get(
        f'{lifecycle_url}/api/v1/fatman/{manifest.name}/logs',
        params={'tail': tail},
        verify=False,
    )
    response = parse_response(r, 'Lifecycle response error')
    logs: str = response['logs']
    log_lines = len(logs.splitlines())
    logging.info(f'Recent logs of "{manifest.name}" fatman ({log_lines} lines):\n')
    print(logs)



import dataclasses
import logging
from typing import Dict

import dacite
from racetrack_client.log.context_error import wrap_context
from racetrack_client.client_config.io import load_client_config, save_client_config
from racetrack_client.client_config.client_config import Credentials, ClientConfig
from racetrack_client.utils.url import trim_url


def set_credentials(repo_url: str, username: str, token_password: str):
    client_config = load_client_config()
    client_config.git_credentials[repo_url] = Credentials(username=username, password=token_password)
    logging.info(f'git credentials added for repo: {repo_url}')
    save_client_config(client_config)


def set_config_setting(setting_name: str, setting_value: str):
    client_config = load_client_config()
    config_dict: Dict = dataclasses.asdict(client_config)

    assert setting_name in config_dict, f'client config doesn\'t have setting named {setting_name}'
    config_dict[setting_name] = setting_value
    with wrap_context('converting setting to target data type'):
        client_config = dacite.from_dict(
            data_class=ClientConfig,
            data=config_dict,
            config=dacite.Config(cast=[]),
        )

    logging.info(f'Client setting {setting_name} set to: {setting_value}')
    save_client_config(client_config)


def set_config_url_alias(alias: str, lifecycle_url: str):
    """Set up an alias for Lifecycle URL"""
    client_config = load_client_config()
    lifecycle_url = trim_url(lifecycle_url)
    client_config.lifecycle_url_aliases[alias] = lifecycle_url
    logging.info(f'Alias "{alias}" set to Racetrack URL {lifecycle_url}')
    save_client_config(client_config)
